"use strict";
/**
 * @fileoverview SIP Transactions
 */

/**
 * SIP Transactions module.
 * @augments SIP
 */
module.exports = function(SIP) {
  var
    C = {
      // Transaction states
      STATUS_TRYING: 1,
      STATUS_PROCEEDING: 2,
      STATUS_CALLING: 3,
      STATUS_ACCEPTED: 4,
      STATUS_COMPLETED: 5,
      STATUS_TERMINATED: 6,
      STATUS_CONFIRMED: 7,

      // Transaction types
      NON_INVITE_CLIENT: 'nict',
      NON_INVITE_SERVER: 'nist',
      INVITE_CLIENT: 'ict',
      INVITE_SERVER: 'ist'
    };

  /**
   * @augments SIP.Transactions
   * @class Non Invite Client Transaction
   * @param {SIP.RequestSender} request_sender
   * @param {SIP.OutgoingRequest} request
   * @param {SIP.Transport} transport
   */
  var NonInviteClientTransaction = function(request_sender, request, transport) {
    var via;

    this.type = C.NON_INVITE_CLIENT;
    this.transport = transport;
    this.id = 'z9hG4bK' + Math.floor(Math.random() * 10000000);
    this.request_sender = request_sender;
    this.request = request;

    // if route header, send to it else send to RURI
    var r = {};
    if (this.request.getHeaders("Route").length > 0) {
      var routeHdr = this.request.getHeaders("Route");
      if (routeHdr[0]) {
        r = routeHdr[0].replace('<', '').replace('>', '');
        // parse this URI
        r = SIP.URI.parse(r);
      }
    } else {

      var ruri = this.request.ruri;
      if (!(this.request.ruri instanceof SIP.URI)) {
        ruri = SIP.URI.parse(this.request.ruri)
      }

      r = ruri;
    }

    if(!r.host) {
      // todo error out properly
      console.log('No route to host')
      return;
    }

    this.host = r.host;

    if (!r.port) {
      this.port = 5060;
    } else {
      this.port = r.port;
    }

    this.logger = request_sender.ua.getLogger('sip.transaction.nict', this.id);

    via = 'SIP/2.0/UDP';
    via += ' ' + request_sender.ua.configuration.viaHost + ':' + request_sender.ua.configuration.viaPort + ';branch=' + this.id;

    this.request.setHeader('via', via);

    this.request_sender.ua.newTransaction(this);
  };
  NonInviteClientTransaction.prototype = Object.create(SIP.EventEmitter.prototype);

  NonInviteClientTransaction.prototype.stateChanged = function(state) {
    this.state = state;
    this.emit('stateChanged');
  };

  NonInviteClientTransaction.prototype.send = function() {
    var tr = this;

    this.stateChanged(C.STATUS_TRYING);
    this.F = SIP.Timers.setTimeout(tr.timer_F.bind(tr), SIP.Timers.TIMER_F);

    if (!this.transport.send(this.host, this.port, this.request)) {
      this.onTransportError();
    }
  };

  NonInviteClientTransaction.prototype.onTransportError = function() {
    this.logger.log('transport error occurred, deleting non-INVITE client transaction ' + this.id);
    SIP.Timers.clearTimeout(this.F);
    SIP.Timers.clearTimeout(this.K);
    this.stateChanged(C.STATUS_TERMINATED);
    this.request_sender.ua.destroyTransaction(this);
    this.request_sender.onTransportError();
  };

  NonInviteClientTransaction.prototype.timer_F = function() {
    this.logger.log('Timer F expired for non-INVITE client transaction ' + this.id);
    this.stateChanged(C.STATUS_TERMINATED);
    this.request_sender.ua.destroyTransaction(this);
    this.request_sender.onRequestTimeout();
  };

  NonInviteClientTransaction.prototype.timer_K = function() {
    this.stateChanged(C.STATUS_TERMINATED);
    this.request_sender.ua.destroyTransaction(this);
  };

  NonInviteClientTransaction.prototype.receiveResponse = function(response) {
    var
      tr = this,
      status_code = response.status_code;

    if (status_code < 200) {
      switch (this.state) {
        case C.STATUS_TRYING:
        case C.STATUS_PROCEEDING:
          this.stateChanged(C.STATUS_PROCEEDING);
          this.request_sender.receiveResponse(response);
          break;
      }
    } else {
      switch (this.state) {
        case C.STATUS_TRYING:
        case C.STATUS_PROCEEDING:
          this.stateChanged(C.STATUS_COMPLETED);
          SIP.Timers.clearTimeout(this.F);

          if (status_code === 408) {
            this.request_sender.onRequestTimeout();
          } else {
            this.request_sender.receiveResponse(response);
          }

          this.K = SIP.Timers.setTimeout(tr.timer_K.bind(tr), SIP.Timers.TIMER_K);
          break;
        case C.STATUS_COMPLETED:
          break;
      }
    }
  };



  /**
   * @augments SIP.Transactions
   * @class Invite Client Transaction
   * @param {SIP.RequestSender} request_sender
   * @param {SIP.OutgoingRequest} request
   * @param {SIP.Transport} transport
   */
  var InviteClientTransaction = function(request_sender, request, transport) {
    var via,
      tr = this;

    this.type = C.INVITE_CLIENT;
    this.transport = transport;
    this.id = 'z9hG4bK' + Math.floor(Math.random() * 10000000);
    this.request_sender = request_sender;
    this.request = request;

    // if route header, send to it else send to RURI
    var r = {};
    if (this.request.getHeaders("Route").length > 0) {
      var routeHdr = this.request.getHeaders("Route");
      if (routeHdr[0]) {
        r = routeHdr[0].replace('<', '').replace('>', '');
        // parse this URI
        r = SIP.URI.parse(r);
      }
    } else {

      var ruri = this.request.ruri;
      if (!(this.request.ruri instanceof SIP.URI)) {
        ruri = SIP.URI.parse(this.request.ruri)
      }

      r = ruri;
    }

    if(!r.host) {
      // todo error out properly
      console.log('No route to host')
      return;
    }

    this.host = r.host;

    if (!r.port) {
      this.port = 5060;
    } else {
      this.port = r.port;
    }

    this.logger = request_sender.ua.getLogger('sip.transaction.ict', this.id);

    via = 'SIP/2.0/UDP';
    via += ' ' + request_sender.ua.configuration.viaHost + ':' + request_sender.ua.configuration.viaPort + ';branch=' + this.id;

    this.request.setHeader('via', via);

    this.request_sender.ua.newTransaction(this);

    // Add the cancel property to the request.
    //Will be called from the request instance, not the transaction itself.
    this.request.cancel = function(reason) {
      tr.cancel_request(tr, reason);
    };
  };
  InviteClientTransaction.prototype = Object.create(SIP.EventEmitter.prototype);

  InviteClientTransaction.prototype.stateChanged = function(state) {
    this.state = state;
    this.emit('stateChanged');
  };

  InviteClientTransaction.prototype.send = function() {
    var tr = this;
    this.stateChanged(C.STATUS_CALLING);
    this.B = SIP.Timers.setTimeout(tr.timer_B.bind(tr), SIP.Timers.TIMER_B);

    if (!this.transport.send(this.host, this.port, this.request)) {
      this.onTransportError();
    }
  };

  InviteClientTransaction.prototype.onTransportError = function() {
    this.logger.log('transport error occurred, deleting INVITE client transaction ' + this.id);
    SIP.Timers.clearTimeout(this.B);
    SIP.Timers.clearTimeout(this.D);
    SIP.Timers.clearTimeout(this.M);
    this.stateChanged(C.STATUS_TERMINATED);
    this.request_sender.ua.destroyTransaction(this);

    if (this.state !== C.STATUS_ACCEPTED) {
      this.request_sender.onTransportError();
    }
  };

  // RFC 6026 7.2
  InviteClientTransaction.prototype.timer_M = function() {
    this.logger.log('Timer M expired for INVITE client transaction ' + this.id);

    if (this.state === C.STATUS_ACCEPTED) {
      SIP.Timers.clearTimeout(this.B);
      this.stateChanged(C.STATUS_TERMINATED);
      this.request_sender.ua.destroyTransaction(this);
    }
  };

  // RFC 3261 17.1.1
  InviteClientTransaction.prototype.timer_B = function() {
    this.logger.log('Timer B expired for INVITE client transaction ' + this.id);
    if (this.state === C.STATUS_CALLING) {
      this.stateChanged(C.STATUS_TERMINATED);
      this.request_sender.ua.destroyTransaction(this);
      this.request_sender.onRequestTimeout();
    }
  };

  InviteClientTransaction.prototype.timer_D = function() {
    this.logger.log('Timer D expired for INVITE client transaction ' + this.id);
    SIP.Timers.clearTimeout(this.B);
    this.stateChanged(C.STATUS_TERMINATED);
    this.request_sender.ua.destroyTransaction(this);
  };

  InviteClientTransaction.prototype.sendACK = function(response) {
    var tr = this;

    this.ack = 'ACK ' + this.request.ruri + ' SIP/2.0\r\n';
    this.ack += 'Via: ' + this.request.headers['Via'].toString() + '\r\n';

    if (this.request.headers['Route']) {
      this.ack += 'Route: ' + this.request.headers['Route'].toString() + '\r\n';
    }

    this.ack += 'To: ' + response.getHeader('to') + '\r\n';
    this.ack += 'From: ' + this.request.headers['From'].toString() + '\r\n';
    this.ack += 'Call-ID: ' + this.request.headers['Call-ID'].toString() + '\r\n';
    this.ack += 'CSeq: ' + this.request.headers['CSeq'].toString().split(' ')[0];
    this.ack += ' ACK\r\n\r\n';

    this.D = SIP.Timers.setTimeout(tr.timer_D.bind(tr), SIP.Timers.TIMER_D);

    this.transport.send(this.host, this.port, this.ack);
  };

  InviteClientTransaction.prototype.cancel_request = function(tr, reason) {
    var request = tr.request;

    this.cancel = SIP.C.CANCEL + ' ' + request.ruri + ' SIP/2.0\r\n';
    this.cancel += 'Via: ' + request.headers['Via'].toString() + '\r\n';

    if (this.request.headers['Route']) {
      this.cancel += 'Route: ' + request.headers['Route'].toString() + '\r\n';
    }

    this.cancel += 'To: ' + request.headers['To'].toString() + '\r\n';
    this.cancel += 'From: ' + request.headers['From'].toString() + '\r\n';
    this.cancel += 'Call-ID: ' + request.headers['Call-ID'].toString() + '\r\n';
    this.cancel += 'CSeq: ' + request.headers['CSeq'].toString().split(' ')[0] +
      ' CANCEL\r\n';

    if (reason) {
      this.cancel += 'Reason: ' + reason + '\r\n';
    }

    this.cancel += 'Content-Length: 0\r\n\r\n';

    // Send only if a provisional response (>100) has been received.
    if (this.state === C.STATUS_PROCEEDING) {
      this.transport.send(this.host, this.port, this.cancel);
    }
  };

  InviteClientTransaction.prototype.receiveResponse = function(response) {
    var
      tr = this,
      status_code = response.status_code;

    if (status_code >= 100 && status_code <= 199) {
      switch (this.state) {
        case C.STATUS_CALLING:
          this.stateChanged(C.STATUS_PROCEEDING);
          this.request_sender.receiveResponse(response);
          if (this.cancel) {
            this.transport.send(this.host, this.port, this.cancel);
          }
          break;
        case C.STATUS_PROCEEDING:
          this.request_sender.receiveResponse(response);
          break;
      }
    } else if (status_code >= 200 && status_code <= 299) {
      switch (this.state) {
        case C.STATUS_CALLING:
        case C.STATUS_PROCEEDING:
          this.stateChanged(C.STATUS_ACCEPTED);
          this.M = SIP.Timers.setTimeout(tr.timer_M.bind(tr), SIP.Timers.TIMER_M);
          this.request_sender.receiveResponse(response);
          break;
        case C.STATUS_ACCEPTED:
          this.request_sender.receiveResponse(response);
          break;
      }
    } else if (status_code >= 300 && status_code <= 699) {
      switch (this.state) {
        case C.STATUS_CALLING:
        case C.STATUS_PROCEEDING:
          this.stateChanged(C.STATUS_COMPLETED);
          this.sendACK(response);
          this.request_sender.receiveResponse(response);
          break;
        case C.STATUS_COMPLETED:
          this.sendACK(response);
          break;
      }
    }
  };


  /**
   * @augments SIP.Transactions
   * @class ACK Client Transaction
   * @param {SIP.RequestSender} request_sender
   * @param {SIP.OutgoingRequest} request
   * @param {SIP.Transport} transport
   */
  var AckClientTransaction = function(request_sender, request, transport) {
    var via;

    this.transport = transport;
    this.id = 'z9hG4bK' + Math.floor(Math.random() * 10000000);
    this.request_sender = request_sender;
    this.request = request;


    // if route header, send to it else send to RURI
    var r = {};
    if (this.request.getHeaders("Route").length > 0) {
      var routeHdr = this.request.getHeaders("Route");
      if (routeHdr[0]) {
        r = routeHdr[0].replace('<', '').replace('>', '');
        // parse this URI
        r = SIP.URI.parse(r);
      }
    } else {

      var ruri = this.request.ruri;
      if (!(this.request.ruri instanceof SIP.URI)) {
        ruri = SIP.URI.parse(this.request.ruri)
      }

      r = ruri;
    }

    if(!r.host) {
      // todo error out properly
      console.log('No route to host')
      return;
    }

    this.host = r.host;

    if (!r.port) {
      this.port = 5060;
    } else {
      this.port = r.port;
    }

    this.logger = request_sender.ua.getLogger('sip.transaction.nict', this.id);

    via = 'SIP/2.0/UDP';
    via += ' ' + request_sender.ua.configuration.viaHost + ':' + request_sender.ua.configuration.viaPort + ';branch=' + this.id;

    this.request.setHeader('via', via);
  };
  AckClientTransaction.prototype = Object.create(SIP.EventEmitter.prototype);

  AckClientTransaction.prototype.send = function() {
    if (!this.transport.send(this.host, this.port, this.request)) {
      this.onTransportError();
    }
  };

  AckClientTransaction.prototype.onTransportError = function() {
    this.logger.log('transport error occurred, for an ACK client transaction ' + this.id);
    this.request_sender.onTransportError();
  };


  /**
   * @augments SIP.Transactions
   * @class Non Invite Server Transaction
   * @param {SIP.IncomingRequest} request
   * @param {SIP.UA} ua
   */
  var NonInviteServerTransaction = function(request, ua) {
    this.type = C.NON_INVITE_SERVER;
    this.host = request.via.host;
    this.port = request.via.port || 5060;
    this.id = request.via_branch;
    this.request = request;
    this.transport = request.transport;
    this.ua = ua;
    this.last_response = '';
    request.server_transaction = this;

    this.logger = ua.getLogger('sip.transaction.nist', this.id);

    this.state = C.STATUS_TRYING;

    ua.newTransaction(this);
  };
  NonInviteServerTransaction.prototype = Object.create(SIP.EventEmitter.prototype);

  NonInviteServerTransaction.prototype.stateChanged = function(state) {
    this.state = state;
    this.emit('stateChanged');
  };

  NonInviteServerTransaction.prototype.timer_J = function() {
    this.logger.log('Timer J expired for non-INVITE server transaction ' + this.id);
    this.stateChanged(C.STATUS_TERMINATED);
    this.ua.destroyTransaction(this);
  };

  NonInviteServerTransaction.prototype.onTransportError = function() {
    if (!this.transportError) {
      this.transportError = true;

      this.logger.log('transport error occurred, deleting non-INVITE server transaction ' + this.id);

      SIP.Timers.clearTimeout(this.J);
      this.stateChanged(C.STATUS_TERMINATED);
      this.ua.destroyTransaction(this);
    }
  };

  NonInviteServerTransaction.prototype.receiveResponse = function(status_code, response) {
    var tr = this;
    var deferred = SIP.Utils.defer();

    if (status_code === 100) {
      /* RFC 4320 4.1
       * 'A SIP element MUST NOT
       * send any provisional response with a
       * Status-Code other than 100 to a non-INVITE request.'
       */
      switch (this.state) {
        case C.STATUS_TRYING:
          this.stateChanged(C.STATUS_PROCEEDING);
          if (!this.transport.send(this.host, this.port, response)) {
            this.onTransportError();
          }
          break;
        case C.STATUS_PROCEEDING:
          this.last_response = response;
          if (!this.transport.send(this.host, this.port, response)) {
            this.onTransportError();
            deferred.reject();
          } else {
            deferred.resolve();
          }
          break;
      }
    } else if (status_code >= 200 && status_code <= 699) {
      switch (this.state) {
        case C.STATUS_TRYING:
        case C.STATUS_PROCEEDING:
          this.stateChanged(C.STATUS_COMPLETED);
          this.last_response = response;
          this.J = SIP.Timers.setTimeout(tr.timer_J.bind(tr), SIP.Timers.TIMER_J);
          if (!this.transport.send(this.host, this.port, response)) {
            this.onTransportError();
            deferred.reject();
          } else {
            deferred.resolve();
          }
          break;
        case C.STATUS_COMPLETED:
          break;
      }
    }

    return deferred.promise;
  };

  /**
   * @augments SIP.Transactions
   * @class Invite Server Transaction
   * @param {SIP.IncomingRequest} request
   * @param {SIP.UA} ua
   */
  var InviteServerTransaction = function(request, ua) {
    this.type = C.INVITE_SERVER;
    this.host = request.via.host;
    this.port = request.via.port || 5060;
    this.id = request.via_branch;
    this.request = request;
    this.transport = request.transport;
    this.ua = ua;
    this.last_response = '';
    request.server_transaction = this;

    this.logger = ua.getLogger('sip.transaction.ist', this.id);

    this.state = C.STATUS_PROCEEDING;

    ua.newTransaction(this);

    this.resendProvisionalTimer = null;

    request.reply(100);
  };
  InviteServerTransaction.prototype = Object.create(SIP.EventEmitter.prototype);

  InviteServerTransaction.prototype.stateChanged = function(state) {
    this.state = state;
    this.emit('stateChanged');
  };

  InviteServerTransaction.prototype.timer_H = function() {
    this.logger.log('Timer H expired for INVITE server transaction ' + this.id);

    if (this.state === C.STATUS_COMPLETED) {
      this.logger.warn('transactions', 'ACK for INVITE server transaction was never received, call will be terminated');
    }

    this.stateChanged(C.STATUS_TERMINATED);
    this.ua.destroyTransaction(this);
  };

  InviteServerTransaction.prototype.timer_I = function() {
    this.stateChanged(C.STATUS_TERMINATED);
    this.ua.destroyTransaction(this);
  };

  // RFC 6026 7.1
  InviteServerTransaction.prototype.timer_L = function() {
    this.logger.log('Timer L expired for INVITE server transaction ' + this.id);

    if (this.state === C.STATUS_ACCEPTED) {
      this.stateChanged(C.STATUS_TERMINATED);
      this.ua.destroyTransaction(this);
    }
  };

  InviteServerTransaction.prototype.onTransportError = function() {
    if (!this.transportError) {
      this.transportError = true;

      this.logger.log('transport error occurred, deleting INVITE server transaction ' + this.id);

      if (this.resendProvisionalTimer !== null) {
        SIP.Timers.clearInterval(this.resendProvisionalTimer);
        this.resendProvisionalTimer = null;
      }

      SIP.Timers.clearTimeout(this.L);
      SIP.Timers.clearTimeout(this.H);
      SIP.Timers.clearTimeout(this.I);

      this.stateChanged(C.STATUS_TERMINATED);
      this.ua.destroyTransaction(this);
    }
  };

  InviteServerTransaction.prototype.resend_provisional = function() {
    if (!this.transport.send(this.host, this.port, this.last_response)) {
      this.onTransportError();
    }
  };

  // INVITE Server Transaction RFC 3261 17.2.1
  InviteServerTransaction.prototype.receiveResponse = function(status_code, response) {
    var tr = this;
    var deferred = SIP.Utils.defer();

    if (status_code >= 100 && status_code <= 199) {
      switch (this.state) {
        case C.STATUS_PROCEEDING:
          if (!this.transport.send(this.host, this.port, response)) {
            this.onTransportError();
          }
          this.last_response = response;
          break;
      }
    }

    if (status_code > 100 && status_code <= 199 && this.state === C.STATUS_PROCEEDING) {
      // Trigger the resendProvisionalTimer only for the first non 100 provisional response.
      if (this.resendProvisionalTimer === null) {
        this.resendProvisionalTimer = SIP.Timers.setInterval(tr.resend_provisional.bind(tr),
          SIP.Timers.PROVISIONAL_RESPONSE_INTERVAL);
      }
    } else if (status_code >= 200 && status_code <= 299) {
      switch (this.state) {
        case C.STATUS_PROCEEDING:
          this.stateChanged(C.STATUS_ACCEPTED);
          this.last_response = response;
          this.L = SIP.Timers.setTimeout(tr.timer_L.bind(tr), SIP.Timers.TIMER_L);

          if (this.resendProvisionalTimer !== null) {
            SIP.Timers.clearInterval(this.resendProvisionalTimer);
            this.resendProvisionalTimer = null;
          }
          /* falls through */
        case C.STATUS_ACCEPTED:
          // Note that this point will be reached for proceeding tr.state also.
          if (!this.transport.send(this.host, this.port, response)) {
            this.onTransportError();
            deferred.reject();
          } else {
            deferred.resolve();
          }
          break;
      }
    } else if (status_code >= 300 && status_code <= 699) {
      switch (this.state) {
        case C.STATUS_PROCEEDING:
          if (this.resendProvisionalTimer !== null) {
            SIP.Timers.clearInterval(this.resendProvisionalTimer);
            this.resendProvisionalTimer = null;
          }

          if (!this.transport.send(this.host, this.port, response)) {
            this.onTransportError();
            deferred.reject();
          } else {
            this.stateChanged(C.STATUS_COMPLETED);
            this.H = SIP.Timers.setTimeout(tr.timer_H.bind(tr), SIP.Timers.TIMER_H);
            deferred.resolve();
          }
          break;
      }
    }

    return deferred.promise;
  };

  /**
   * @function
   * @param {SIP.UA} ua
   * @param {SIP.IncomingRequest} request
   *
   * @return {boolean}
   * INVITE:
   *  _true_ if retransmission
   *  _false_ new request
   *
   * ACK:
   *  _true_  ACK to non2xx response
   *  _false_ ACK must be passed to TU (accepted state)
   *          ACK to 2xx response
   *
   * CANCEL:
   *  _true_  no matching invite transaction
   *  _false_ matching invite transaction and no final response sent
   *
   * OTHER:
   *  _true_  retransmission
   *  _false_ new request
   */
  var checkTransaction = function(ua, request) {
    var tr;

    switch (request.method) {
      case SIP.C.INVITE:
        tr = ua.transactions.ist[request.via_branch];
        if (tr) {
          switch (tr.state) {
            case C.STATUS_PROCEEDING:
              tr.transport.send(tr.host, tr.port, tr.last_response);
              break;

              // RFC 6026 7.1 Invite retransmission
              //received while in C.STATUS_ACCEPTED state. Absorb it.
            case C.STATUS_ACCEPTED:
              break;
          }
          return true;
        }
        break;
      case SIP.C.ACK:
        tr = ua.transactions.ist[request.via_branch];

        // RFC 6026 7.1
        if (tr) {
          if (tr.state === C.STATUS_ACCEPTED) {
            return false;
          } else if (tr.state === C.STATUS_COMPLETED) {
            tr.state = C.STATUS_CONFIRMED;
            tr.I = SIP.Timers.setTimeout(tr.timer_I.bind(tr), SIP.Timers.TIMER_I);
            return true;
          }
        }

        // ACK to 2XX Response.
        else {
          return false;
        }
        break;
      case SIP.C.CANCEL:
        tr = ua.transactions.ist[request.via_branch];
        if (tr) {
          request.reply_sl(200);
          if (tr.state === C.STATUS_PROCEEDING) {
            return false;
          } else {
            return true;
          }
        } else {
          request.reply_sl(481);
          return true;
        }
        break;
      default:

        // Non-INVITE Server Transaction RFC 3261 17.2.2
        tr = ua.transactions.nist[request.via_branch];
        if (tr) {
          switch (tr.state) {
            case C.STATUS_TRYING:
              break;
            case C.STATUS_PROCEEDING:
            case C.STATUS_COMPLETED:
              tr.transport.send(tr.host, tr.port, tr.last_response);
              break;
          }
          return true;
        }
        break;
    }
  };

  SIP.Transactions = {
    C: C,
    checkTransaction: checkTransaction,
    NonInviteClientTransaction: NonInviteClientTransaction,
    InviteClientTransaction: InviteClientTransaction,
    AckClientTransaction: AckClientTransaction,
    NonInviteServerTransaction: NonInviteServerTransaction,
    InviteServerTransaction: InviteServerTransaction
  };

};
